import math
import numpy

from PyQt5 import QtCore, QtGui

from API import BuilderPanelPlugin
from QuantumATK import Angstrom, BulkConfiguration

from NL.GUI.Presenter.Presenter import Presenter
from NL.CommonConcepts.Configurations.ConfigurationCopy import configurationCopy
from NL.Math.Utilities import rotationMatrix

from .UI.TwisterWidget import TwisterWidget

def twist(x, rotation_angle_per_z, rotation_axis,
          rotation_axis_center, z_start, z_end):
    """
    Function for twisting a 1-d structure
    @param x : Coordinates of 1-d structure
    @param rotation_angle_per_z : size of twist in angle/length
    @param rotation_axis : axis to apply twist along
    @param rotation_axis_center : center of the rotation axis
    @param z_start : z value for starting the twist
    @param z_end : z value for ending the twist
    """

    # do not twist for z > z_end
    z = x[2]
    z = min(z,z_end)
    # do not twist for z < z_start
    z = z - z_start
    z = max (z,0.0)
    # find twist angle
    theta = z*rotation_angle_per_z
    # calculate the rotation matrix
    rotation_matrix = rotationMatrix(theta, *rotation_axis)
    # apply rotation
    return rotation_axis_center + numpy.dot(rotation_matrix, x - rotation_axis_center)

class TwisterPlugin(BuilderPanelPlugin):
    """
       Plugin for buckling
    """
    def __init__(self, parent=None):
        """
        Constructor.

        @parent : The widget that holds the plugin.
        """
        self._widget = TwisterWidget()

        BuilderPanelPlugin.__init__(self, parent)

        # Placeholder for the preview
        self._preview = None

        self._widget.apply_button.clicked.connect(self.applyClicked)
        self._widget.preview_button.clicked.connect(self.preview)
        self._widget.angle.valueChanged[float].connect(self.preview)
        self._widget.nontwistingzone.valueChanged[float].connect(self.preview)

    def title(self):
        """
        @return The title of the panel.
        """
        return self.tr("Twister")

    def widget(self):
        """
        @return The widget presented to the user.
        """
        return self._widget

    @staticmethod
    def group():
        return 'Coordinate Tools'

    def doTwisting(self):
        """
        Retuns the twisted coordinates
        """
        # Get active presenter and configuration
        presenter = self.stash().activePresenter()
        configuration = presenter.configuration()

        # Parameters from the widget
        twist_angle = self._widget.angle.value()
        nontwistingzone = self._widget.nontwistingzone.value()

        # Properties of the bulk configuration
        lattice = configuration.bravaisLattice()
        cartesian_coordinates = configuration.cartesianCoordinates().inUnitsOf(Angstrom)

        # calculate the length of the 1-d structure
        z_length = numpy.linalg.norm(lattice.primitiveVectors()[2].inUnitsOf(Angstrom))

        # calculate twist parameters
        rotation_angle_per_z = twist_angle/180. * math.pi /(z_length-2*nontwistingzone)
        rotation_axis = numpy.array([0,0,1])
        rotation_axis_center = numpy.sum(cartesian_coordinates,axis=0)/len(cartesian_coordinates)

        # define a function of one variable, f(c), for twisting the atoms
        f = lambda c : twist(c, rotation_angle_per_z, rotation_axis,
                             rotation_axis_center, nontwistingzone,z_length-nontwistingzone)
        # apply the function to find new displaced coordinates
        twisted_coordinates = numpy.apply_along_axis(f, 1, cartesian_coordinates)

        return twisted_coordinates

    def applyClicked(self):
        """
        Called when the 'Apply' button is pressed.
        """
        # Get active presenter and configuration
        presenter = self.stash().activePresenter()
        configuration = presenter.configuration()

        # Get buckled coordinates
        new_coordinates = self.doTwisting()

        # Set undo, then give new config to presenter
        presenter.setUndoMark('Twist')
        presenter.changeAtoms(positions=new_coordinates*Angstrom)

        # Remove preview
        self.stash().setPreview(None, self)

    def preview(self):
        """
        Called when the 'Preview' button is pressed or any value is changed.
        """
        # Get buckled coordinates
        new_coordinates = self.doTwisting()

        self._preview.changeAtoms(positions=new_coordinates*Angstrom)
        self.stash().setPreview(self._preview, self)

    def activeConfigurationModified(self):
        """
           Called when changes occur in the stash of configurations.
        """
        # Set enabled state based on configuration
        configuration = self.stash().activeConfiguration()
        if isinstance(configuration, BulkConfiguration) and len(configuration) > 0:
            self.setAvailable(True)
            # Make a local copy of the configuration to use for the preview
            if self._preview == None:
                self._preview = Presenter(configurationCopy(configuration))
            else:
                self._preview.setConfiguration(configurationCopy(configuration))
        else:
            self.setAvailable(False)
